@@ -280,6 +280,145 @@ GST_START_TEST (test_event)
 
 GST_END_TEST;
 
+typedef struct _ProbeEvent
+{
+  gboolean received;
+  gdouble x_pos, y_pos;
+} ProbeEvent;
+
+static GstPadProbeReturn
+probe_nav_event (GstPad * pad, GstPadProbeInfo * info, gpointer user_data)
+{
+  GstEvent *event = gst_pad_probe_info_get_event (info);
+  ProbeEvent *probe_ev = (ProbeEvent *) user_data;
+
+  if (GST_EVENT_TYPE (event) == GST_EVENT_NAVIGATION) {
+    probe_ev->received = TRUE;
+    gst_navigation_event_parse_mouse_move_event (event,
+        &(probe_ev->x_pos), &(probe_ev->y_pos));
+  }
+
+  return GST_PAD_PROBE_OK;
+}
+
+GST_START_TEST (test_navigation_events)
+{
+  GstElement *bin, *src1, *src2, *src3, *filter1, *filter2, *filter3;
+  GstElement *compositor, *sink;
+  GstPad *srcpad, *sinkpad;
+  GstCaps *caps;
+  gboolean res;
+  ProbeEvent probe_events[3];
+  GstStateChangeReturn state_res;
+  GstEvent *event;
+
+  GST_INFO ("preparing test");
+
+  /* build pipeline */
+  bin = gst_pipeline_new ("pipeline");
+  src1 = gst_element_factory_make ("videotestsrc", "src1");
+  src2 = gst_element_factory_make ("videotestsrc", "src2");
+  src3 = gst_element_factory_make ("videotestsrc", "src3");
+  filter1 = gst_element_factory_make ("capsfilter", "filter1");
+  filter2 = gst_element_factory_make ("capsfilter", "filter2");
+  filter3 = gst_element_factory_make ("capsfilter", "filter3");
+  compositor = gst_element_factory_make ("compositor", "compositor");
+  sink = gst_element_factory_make ("fakesink", "sink");
+  gst_bin_add_many (GST_BIN (bin), src1, src2, src3, filter1, filter2, filter3,
+      compositor, sink, NULL);
+
+  /* configure capsfilters */
+  caps = gst_caps_from_string ("video/x-raw,width=800,height=400");
+  g_object_set (filter1, "caps", caps, NULL);
+  gst_caps_unref (caps);
+  caps = gst_caps_from_string ("video/x-raw,width=400,height=200");
+  g_object_set (filter2, "caps", caps, NULL);
+  gst_caps_unref (caps);
+  caps = gst_caps_from_string ("video/x-raw,width=200,height=50");
+  g_object_set (filter3, "caps", caps, NULL);
+  gst_caps_unref (caps);
+
+  res = gst_element_link_many (src1, filter1, compositor, NULL);
+  fail_unless (res == TRUE, NULL);
+  res = gst_element_link_many (src2, filter2, compositor, NULL);
+  fail_unless (res == TRUE, NULL);
+  res = gst_element_link_many (src3, filter3, compositor, NULL);
+  fail_unless (res == TRUE, NULL);
+  res = gst_element_link (compositor, sink);
+  fail_unless (res == TRUE, NULL);
+
+  srcpad = gst_element_get_static_pad (compositor, "src");
+  gst_object_unref (srcpad);
+
+  /* configure pads and add probes */
+  srcpad = gst_element_get_static_pad (filter1, "src");
+  sinkpad = gst_pad_get_peer (srcpad);
+  g_object_set (sinkpad,
+      "width", 400, "height", 300, "xpos", 200, "ypos", 100, NULL);
+  probe_events[0].received = FALSE;
+  gst_pad_add_probe (srcpad, GST_PAD_PROBE_TYPE_EVENT_UPSTREAM,
+      probe_nav_event, (gpointer) probe_events, NULL);
+  gst_object_unref (srcpad);
+  gst_object_unref (sinkpad);
+
+  srcpad = gst_element_get_static_pad (filter2, "src");
+  sinkpad = gst_pad_get_peer (srcpad);
+  g_object_set (sinkpad,
+      "width", 400, "height", 200, "xpos", 20, "ypos", 0, NULL);
+  probe_events[1].received = FALSE;
+  gst_pad_add_probe (srcpad, GST_PAD_PROBE_TYPE_EVENT_UPSTREAM,
+      probe_nav_event, (gpointer) (probe_events + 1), NULL);
+  gst_object_unref (srcpad);
+  gst_object_unref (sinkpad);
+
+  srcpad = gst_element_get_static_pad (filter3, "src");
+  sinkpad = gst_pad_get_peer (srcpad);
+  g_object_set (sinkpad,
+      "width", 200, "height", 50, "xpos", 0, "ypos", 0, NULL);
+  probe_events[2].received = FALSE;
+  gst_pad_add_probe (srcpad, GST_PAD_PROBE_TYPE_EVENT_UPSTREAM,
+      probe_nav_event, (gpointer) (probe_events + 2), NULL);
+  gst_object_unref (srcpad);
+  gst_object_unref (sinkpad);
+
+  event =
+      gst_event_new_navigation (gst_structure_new
+      ("application/x-gst-navigation", "event", G_TYPE_STRING, "mouse-move",
+          "button", G_TYPE_INT, 0, "pointer_x", G_TYPE_DOUBLE, 350.0,
+          "pointer_y", G_TYPE_DOUBLE, 100.0, NULL));
+
+  GST_INFO ("starting test");
+
+  /* prepare playing */
+  state_res = gst_element_set_state (bin, GST_STATE_PAUSED);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  /* wait for completion */
+  state_res = gst_element_get_state (bin, NULL, NULL, GST_CLOCK_TIME_NONE);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  /* send event and validate */
+  res = gst_element_send_event (sink, event);
+  fail_unless (res == TRUE, NULL);
+
+  /* check received events */
+  ck_assert_msg (probe_events[0].received);
+  ck_assert_msg (probe_events[1].received);
+  ck_assert_msg (!probe_events[2].received);
+
+  ck_assert_int_eq ((gint) probe_events[0].x_pos, 300);
+  ck_assert_int_eq ((gint) probe_events[0].y_pos, 0);
+  ck_assert_int_eq ((gint) probe_events[1].x_pos, 330);
+  ck_assert_int_eq ((gint) probe_events[1].y_pos, 100);
+
+  state_res = gst_element_set_state (bin, GST_STATE_NULL);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  gst_object_unref (bin);
+}
+
+GST_END_TEST;
+
 static GstBuffer *
 create_video_buffer (GstCaps * caps, gint ts_in_seconds)
 {
@@ -1942,7 +2081,7 @@ GST_START_TEST (test_pad_z_order)
 
 GST_END_TEST;
 
-/* 
+/*
  * Test that the pad numbering assigned by aggregator behaves as follows:
  * 1. If a pad number is requested, it must be assigned if it is available
  * 2. When numbering automatically, the largest available pad number is used
@@ -2268,6 +2407,344 @@ GST_START_TEST (test_signals)
 
 GST_END_TEST;
 
+static void
+on_reverse_handoff (GstElement * sink, GstBuffer * buffer, GstPad * pad,
+    GstClockTime * pos)
+{
+  GstClockTime pts = GST_BUFFER_PTS (buffer);
+  GstClockTime dur = GST_BUFFER_DURATION (buffer);
+
+  fail_unless (GST_CLOCK_TIME_IS_VALID (pts));
+  fail_unless_equals_clocktime (dur, GST_MSECOND * 100);
+
+  if (!GST_CLOCK_TIME_IS_VALID (*pos)) {
+    *pos = pts;
+  } else {
+    fail_unless (pts < *pos);
+    *pos = pts;
+  }
+}
+
+GST_START_TEST (test_reverse)
+{
+  GstElement *bin, *src1, *src2, *compositor, *sink;
+  GstElement *cp1, *cp2, *cp3;
+  GstCaps *caps;
+  GstBus *bus;
+  GstEvent *seek_event;
+  GstStateChangeReturn state_res;
+  gboolean res;
+  GstClockTime pos = GST_CLOCK_TIME_NONE;
+
+  GST_INFO ("preparing test");
+
+  /* build pipeline */
+  bin = gst_pipeline_new ("pipeline");
+  bus = gst_element_get_bus (bin);
+  gst_bus_add_signal_watch_full (bus, G_PRIORITY_HIGH);
+
+  src1 = gst_element_factory_make ("videotestsrc", "src1");
+  src2 = gst_element_factory_make ("videotestsrc", "src2");
+  compositor = gst_element_factory_make ("compositor", "compositor");
+  cp1 = gst_element_factory_make ("capsfilter", "cp1");
+  cp2 = gst_element_factory_make ("capsfilter", "cp2");
+  cp3 = gst_element_factory_make ("capsfilter", "cp3");
+  sink = gst_element_factory_make ("fakesink", "sink");
+  gst_bin_add_many (GST_BIN (bin), src1, src2, compositor, sink, cp1, cp2,
+      cp3, NULL);
+
+  res = gst_element_link_many (src1, cp1, compositor, NULL);
+  fail_unless (res == TRUE, NULL);
+  res = gst_element_link_many (src2, cp2, compositor, NULL);
+  fail_unless (res == TRUE, NULL);
+  res = gst_element_link_many (compositor, cp3, sink, NULL);
+  fail_unless (res == TRUE, NULL);
+
+  caps = gst_caps_from_string ("video/x-raw,width=(int)64,height=(int)64,"
+      "framerate=(fraction)10/1");
+  fail_unless (caps != NULL, NULL);
+
+  g_object_set (cp1, "caps", caps, NULL);
+  g_object_set (cp2, "caps", caps, NULL);
+  g_object_set (cp3, "caps", caps, NULL);
+  gst_caps_unref (caps);
+
+  seek_event = gst_event_new_seek (-1.0, GST_FORMAT_TIME,
+      GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_TRICKMODE,
+      GST_SEEK_TYPE_SET, (GstClockTime) 0,
+      GST_SEEK_TYPE_SET, (GstClockTime) 2 * GST_SECOND);
+
+  main_loop = g_main_loop_new (NULL, FALSE);
+  g_signal_connect (bus, "message::error", (GCallback) message_received, bin);
+  g_signal_connect (bus, "message::warning", (GCallback) message_received, bin);
+  g_signal_connect (bus, "message::eos", (GCallback) message_received, bin);
+
+  GST_INFO ("starting test");
+
+  /* prepare playing */
+  state_res = gst_element_set_state (bin, GST_STATE_PAUSED);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  /* wait for completion */
+  state_res = gst_element_get_state (bin, NULL, NULL, GST_CLOCK_TIME_NONE);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  g_object_set (sink, "signal-handoffs", TRUE, NULL);
+  g_signal_connect (sink, "handoff", G_CALLBACK (on_reverse_handoff), &pos);
+
+  res = gst_element_send_event (bin, seek_event);
+  fail_unless (res == TRUE, NULL);
+
+  /* run pipeline */
+  state_res = gst_element_set_state (bin, GST_STATE_PLAYING);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  GST_INFO ("running main loop");
+  g_main_loop_run (main_loop);
+
+  state_res = gst_element_set_state (bin, GST_STATE_NULL);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  fail_unless_equals_clocktime (pos, 0);
+
+  /* cleanup */
+  g_main_loop_unref (main_loop);
+  gst_bus_remove_signal_watch (bus);
+  gst_object_unref (bus);
+  gst_object_unref (bin);
+}
+
+GST_END_TEST;
+
+static void
+eos_test_msg_cb (GstBus * bus, GstMessage * message, gboolean * got_eos)
+{
+  switch (message->type) {
+    case GST_MESSAGE_EOS:
+      *got_eos = TRUE;
+      g_main_loop_quit (main_loop);
+      break;
+    case GST_MESSAGE_ERROR:
+      g_main_loop_quit (main_loop);
+      break;
+    default:
+      break;
+  }
+}
+
+typedef struct
+{
+  GMutex lock;
+  GCond cond;
+  GstBuffer *buf;
+} EosTestData;
+
+static void
+after_eos_handoff_buffer_cb (GstElement * sink, GstBuffer * buf, GstPad * pad,
+    EosTestData * data)
+{
+  g_mutex_lock (&data->lock);
+  gst_buffer_replace (&data->buf, buf);
+  g_cond_signal (&data->cond);
+  g_mutex_unlock (&data->lock);
+}
+
+/* This test checks that aggregator can restart after outputting EOS if
+ * one of its input streams does so.
+ */
+GST_START_TEST (test_stream_start_after_eos)
+{
+  GstElement *bin, *src1, *src2, *compositor, *sink;
+  GstElement *cp1, *cp2, *cp3;
+  GstCaps *caps;
+  GstBus *bus;
+  GstStateChangeReturn state_res;
+  gboolean res;
+  gboolean got_eos = FALSE;
+  EosTestData test_data;
+
+  g_mutex_init (&test_data.lock);
+  g_cond_init (&test_data.cond);
+  test_data.buf = NULL;
+
+  bin = gst_pipeline_new ("pipeline");
+  bus = gst_element_get_bus (bin);
+  gst_bus_add_signal_watch_full (bus, G_PRIORITY_HIGH);
+
+  src1 = gst_element_factory_make ("videotestsrc", NULL);
+  src2 = gst_element_factory_make ("videotestsrc", NULL);
+  compositor = gst_element_factory_make ("compositor", NULL);
+  cp1 = gst_element_factory_make ("capsfilter", NULL);
+  cp2 = gst_element_factory_make ("capsfilter", NULL);
+  cp3 = gst_element_factory_make ("capsfilter", NULL);
+  sink = gst_element_factory_make ("fakesink", NULL);
+  gst_bin_add_many (GST_BIN (bin), src1, src2, compositor, sink, cp1, cp2,
+      cp3, NULL);
+
+  g_object_set (src1, "num-buffers", 1, "is-live", TRUE, NULL);
+  g_object_set (src2, "num-buffers", 1, "is-live", TRUE, NULL);
+
+  res = gst_element_link_many (src1, cp1, compositor, NULL);
+  fail_unless (res);
+  res = gst_element_link_many (src2, cp2, compositor, NULL);
+  fail_unless (res);
+  res = gst_element_link_many (compositor, cp3, sink, NULL);
+  fail_unless (res);
+
+  caps = gst_caps_from_string ("video/x-raw,width=(int)64,height=(int)64,"
+      "framerate=(fraction)10/1");
+  fail_unless (caps != NULL);
+
+  g_object_set (cp1, "caps", caps, NULL);
+  g_object_set (cp2, "caps", caps, NULL);
+  g_object_set (cp3, "caps", caps, NULL);
+  gst_caps_unref (caps);
+
+  main_loop = g_main_loop_new (NULL, FALSE);
+  g_signal_connect (bus, "message::error", (GCallback) eos_test_msg_cb,
+      &got_eos);
+  g_signal_connect (bus, "message::eos", (GCallback) eos_test_msg_cb, &got_eos);
+
+  state_res = gst_element_set_state (bin, GST_STATE_PLAYING);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  g_main_loop_run (main_loop);
+
+  /* Wait for EOS */
+  fail_unless (got_eos);
+
+  /* Configure new input */
+  g_object_set (sink, "signal-handoffs", TRUE, NULL);
+  g_signal_connect (sink, "handoff", (GCallback) after_eos_handoff_buffer_cb,
+      &test_data);
+
+  /* Run a source again */
+  state_res = gst_element_set_state (src1, GST_STATE_NULL);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+  gst_element_sync_state_with_parent (src1);
+
+  g_mutex_lock (&test_data.lock);
+  while (!test_data.buf)
+    g_cond_wait (&test_data.cond, &test_data.lock);
+  g_mutex_unlock (&test_data.lock);
+
+  state_res = gst_element_set_state (bin, GST_STATE_NULL);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  /* cleanup */
+  gst_buffer_unref (test_data.buf);
+  g_main_loop_unref (main_loop);
+  gst_bus_remove_signal_watch (bus);
+  gst_object_unref (bus);
+  gst_object_unref (bin);
+
+  g_mutex_clear (&test_data.lock);
+  g_cond_clear (&test_data.cond);
+}
+
+GST_END_TEST;
+
+/* This test checks that aggregator restarts after outputting an EOS
+ * when a new input stream is added
+ */
+GST_START_TEST (test_new_pad_after_eos)
+{
+  GstElement *bin, *src1, *src2, *src3, *compositor, *sink;
+  GstElement *cp1, *cp2, *cp3, *cp4;
+  GstCaps *caps;
+  GstBus *bus;
+  GstStateChangeReturn state_res;
+  gboolean res;
+  gboolean got_eos = FALSE;
+  EosTestData test_data;
+
+  g_mutex_init (&test_data.lock);
+  g_cond_init (&test_data.cond);
+  test_data.buf = NULL;
+
+  bin = gst_pipeline_new ("pipeline");
+  bus = gst_element_get_bus (bin);
+  gst_bus_add_signal_watch_full (bus, G_PRIORITY_HIGH);
+
+  src1 = gst_element_factory_make ("videotestsrc", NULL);
+  src2 = gst_element_factory_make ("videotestsrc", NULL);
+  src3 = gst_element_factory_make ("videotestsrc", NULL);
+  compositor = gst_element_factory_make ("compositor", NULL);
+  cp1 = gst_element_factory_make ("capsfilter", NULL);
+  cp2 = gst_element_factory_make ("capsfilter", NULL);
+  cp3 = gst_element_factory_make ("capsfilter", NULL);
+  cp4 = gst_element_factory_make ("capsfilter", NULL);
+  sink = gst_element_factory_make ("fakesink", NULL);
+  gst_bin_add_many (GST_BIN (bin), src1, src2, compositor, sink, cp1, cp2,
+      cp3, NULL);
+
+  g_object_set (src1, "num-buffers", 1, "is-live", TRUE, NULL);
+  g_object_set (src2, "num-buffers", 1, "is-live", TRUE, NULL);
+  g_object_set (src3, "num-buffers", 1, "is-live", TRUE, NULL);
+
+  res = gst_element_link_many (src1, cp1, compositor, NULL);
+  fail_unless (res);
+  res = gst_element_link_many (src2, cp2, compositor, NULL);
+  fail_unless (res);
+  res = gst_element_link_many (compositor, cp3, sink, NULL);
+  fail_unless (res);
+
+  caps = gst_caps_from_string ("video/x-raw,width=(int)64,height=(int)64,"
+      "framerate=(fraction)10/1");
+  fail_unless (caps != NULL);
+
+  g_object_set (cp1, "caps", caps, NULL);
+  g_object_set (cp2, "caps", caps, NULL);
+  g_object_set (cp3, "caps", caps, NULL);
+  g_object_set (cp4, "caps", caps, NULL);
+  gst_caps_unref (caps);
+
+  main_loop = g_main_loop_new (NULL, FALSE);
+  g_signal_connect (bus, "message::error", (GCallback) eos_test_msg_cb,
+      &got_eos);
+  g_signal_connect (bus, "message::eos", (GCallback) eos_test_msg_cb, &got_eos);
+
+  state_res = gst_element_set_state (bin, GST_STATE_PLAYING);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  g_main_loop_run (main_loop);
+
+  /* Wait for EOS */
+  fail_unless (got_eos);
+
+  /* Configure new input */
+  g_object_set (sink, "signal-handoffs", TRUE, NULL);
+  g_signal_connect (sink, "handoff", (GCallback) after_eos_handoff_buffer_cb,
+      &test_data);
+
+  gst_bin_add_many (GST_BIN (bin), src3, cp4, NULL);
+  res = gst_element_link_many (src3, cp4, compositor, NULL);
+  fail_unless (res);
+
+  gst_element_sync_state_with_parent (cp4);
+  gst_element_sync_state_with_parent (src3);
+
+  g_mutex_lock (&test_data.lock);
+  while (!test_data.buf)
+    g_cond_wait (&test_data.cond, &test_data.lock);
+  g_mutex_unlock (&test_data.lock);
+
+  state_res = gst_element_set_state (bin, GST_STATE_NULL);
+  ck_assert_int_ne (state_res, GST_STATE_CHANGE_FAILURE);
+
+  /* cleanup */
+  gst_buffer_unref (test_data.buf);
+  g_main_loop_unref (main_loop);
+  gst_bus_remove_signal_watch (bus);
+  gst_object_unref (bus);
+  gst_object_unref (bin);
+
+  g_mutex_clear (&test_data.lock);
+  g_cond_clear (&test_data.cond);
+}
+
+GST_END_TEST;
+
 static Suite *
 compositor_suite (void)
 {
@@ -2277,6 +2754,7 @@ compositor_suite (void)
   suite_add_tcase (s, tc_chain);
   tcase_add_test (tc_chain, test_caps);
   tcase_add_test (tc_chain, test_event);
+  tcase_add_test (tc_chain, test_navigation_events);
   tcase_add_test (tc_chain, test_caps_query);
   tcase_add_test (tc_chain, test_caps_query_interlaced);
   tcase_add_test (tc_chain, test_late_caps_query);
@@ -2308,6 +2786,9 @@ compositor_suite (void)
   tcase_add_test (tc_chain, test_start_time_first_live_drop_3_unlinked_1);
   tcase_add_test (tc_chain, test_gap_events);
   tcase_add_test (tc_chain, test_signals);
+  tcase_add_test (tc_chain, test_reverse);
+  tcase_add_test (tc_chain, test_stream_start_after_eos);
+  tcase_add_test (tc_chain, test_new_pad_after_eos);
 
   return s;
 }
